/*
 * Portable Perl launcher for Windows
 *
 * Approach: Run the Perl interpreter from perl5xx.dll and:
 * - Insert a script name as parameter
 * - Optionally insert include directories as parameter(s)
 *
 * The PAR Packager pp can be used to collect all dependencies of the Perl application
 * Use only the contents of the "lib" directory plus the script from the "script" directory.
 *
 * Inspired by "runperl.c" used to compile the original perl.exe
 *
 * To be linked with "lib/CORE/libperl5xx.a" from your Perl distribution
 * perl5xx.dll needs to be placed in the same directory as the exe
 *
 * 20190516 - proof of concept, request for comments
 * 20190518 - catch strrchr not found, add "perl.exe mode"
 * 20190519 - lint-safe
 *
 * Possible extensions:
 * - set additional lib search path(s)
 * - set an environment variable to let the script know the calling method
 *
 * Author: Oliver Betz https://OliverBetz.de
 * License: Public domain, no liability, no support (useful feedback welcome)
 */

#ifdef _MSC_VER
#include <crtdbg.h> // Copied from runperl.c without further check
#endif

#ifdef __GNUC__
int _CRT_glob = 0; // Why the hell should a compiler glob command line arguments?
#endif

#include <stdlib.h> ///< setenv
#include "EXTERN.h" // from "lib/CORE" of your Perl distribution
#include "perl.h"   // from "lib/CORE" of your Perl distribution

// Three flavours of mangling the path to the exe can be chosen
// by the format of the pathreplace[] string:
// - "\\myscript.pl" makes a fixed script name, independent of the exe name
// - ".pl" replaces just the extension from .exe to .pl
// - "." removes the .exe extension
// The latter two make an "universal" exe (rename it to match the script name)
// No safety belts - don't omit "\\" or "." at the beginning of the string!

static const char pathreplace[] = ".pl"; ///< define the path conversion to the Perl script

// act like the original "perl.exe" if our name is "perl.exe" (case sensitive!)
// To disable this "magic", set to NULL
static const char perlexe[] = "perl.exe"; // set to NULL to disable this "magic"

#define ARGS_ADDED 1 ///< number of arguments we put in front of the user provided args
//#define DEBUGOUT stderr ///< debug output. Comment to disable debug output
#define BUFLEN 2000 ///< \literal PATH_MAX is obsolete since Win10?

int main(int argc, char **argv, char **env)
{
    char buf[BUFLEN]; ///< to construct script path from exe path
    char *cp;
    int i, emulate_perlexe;

// Copied from runperl.c without further check:
#ifdef _MSC_VER
    int currentFlag = _CrtSetDbgFlag(_CRTDBG_REPORT_FLAG);
    currentFlag |= _CRTDBG_LEAK_CHECK_DF;
    _CrtSetDbgFlag(currentFlag);
    _CrtSetBreakAlloc(-1L);
#endif

    buf[0]=0; // ensure a null terminated string
    i = GetModuleFileName(NULL, buf, sizeof(buf));
    if ((i > 0) && (i < ((int)(sizeof buf)-(int)(sizeof pathreplace)-(int)1))){
        buf[sizeof(buf)-1] = 0; // limit strrchr search in case of errors (paranoia)

        cp = strrchr(buf, '\\'); // find the last backslash in path to get the file name

        emulate_perlexe = (cp && perlexe &&  (!strncmp(cp+1, perlexe, sizeof perlexe))); //lint !e506

        cp = strrchr(buf, pathreplace[0]); // find the last delimiter in path
        if(!cp){
            (void)fprintf(stderr, "Failed to find '%c' in %s\n", pathreplace[0], buf);
            return 1; // ---> early return
        }
        if((pathreplace[0] == '.') && (pathreplace[1] == 0)){
            *cp = 0; // Perl script without extension, drop the '.'
        } else {
            (void)memmove(cp, pathreplace, sizeof pathreplace); // paste replacement
        }
    }
    else {
        (void)fprintf(stderr, "Path to %s is too long for my %i bytes buffer \n", argv[0], sizeof(buf));
        return 1; // ---> early return
    }

#ifdef DEBUGOUT
    fprintf(DEBUGOUT, "%i argv parameters were passed to the exe:\n", argc);
    for(i = 0; i < argc; i++){
        fprintf(DEBUGOUT, "%i:%s\n", i, argv[i]);
    }
    if(!emulate_perlexe) fprintf(DEBUGOUT, "\nNow we run the Perl script \"%s\":\n\n", buf);
    else                 fprintf(DEBUGOUT, "\nWe emulate the original perl.exe\n");
#endif

    if(emulate_perlexe){
        i = RunPerl(argc, argv, env); // emulate the standard perl.exe
    } else {
        // generate "our" argument list with additional entries and terminating NULL pointer
        char **ourargv = (char **)malloc((argc+ARGS_ADDED+1) * sizeof(char**));
        if (!ourargv) { (void)fprintf(stderr, "Out of memory building new arg list\n"); return 1; } // ---> early return

        ourargv[0] = argv[0]; // keep filename, although it seems to be dropped by perl
        ourargv[1] = buf; // pass script path to Perl interpreter

        /* copy the remaining user provided arguments and the terminating NULL pointer */
        for (i=1; i<=argc; ++i) {
            ourargv[i+ARGS_ADDED] = argv[i];
        }

        i = RunPerl(argc+ARGS_ADDED, ourargv, env);
        free(ourargv);
    }
    return i; // return code from Perl interpreter like original perl.exe does
}

